package salsa.messaging;

import salsa.language.Actor;
import salsa.naming.UAL;
import salsa.naming.UAN;

/**
	This Interface represents the most basic services a theater must provide a salsa program.
	
	@author stepha
*/

public interface TheaterService {
	/**
		This method returns an actorstate, given a specific UAN.
		If the actorstate for a given UAN is not local to the theater,
		it will return the UAL of that actor
	*/
	public Object getEntry(UAN uan);
	public Object getEntry(UAL ual);

	/**
		This method adds an actorstate to the theaters registry of 
		states.
	*/
	public void setEntry(UAN uan, Object entry);
	public void setEntry(UAL ual, Object entry);

	/**
		These methods remove actors from the registry so that they can
		be picked up by garbage collection - after the die method has been
		invoked.
	*/
	public void removeEntry(UAN uan);
	public void removeEntry(UAL ual);


	/**
		This is used to find the location of the theater
	*/
	public String getLocation();

	/**
		This is used for on-the-fly theater creation,
		it creates a UAL for an anonymous actor.
	*/
	public UAL generateUAL();

	/**
		This is used for identifying tokens.
	*/
	public String getUniqueTokenId();

	/**
		This method returns an object implementing TheaterService. 
		Used for on-the-fly theater startup. (The theater instance might exist but it may only provide
		passive serivces and not listen for incoming actors or messages.) The method will update the 
		actor's UAL after theater startup is successful. If this method is called at the theater already exists,
		the inital actor is registered with the existing theater if it is specified.
		
		@param port The port which the theater should start listening on. 0 indicates any port.
		@param initial Reference to the actor requesting theater startup. 
	*/
	public TheaterService startTheater( Actor initial );
 

	/**
		This method returns the port that the theater is currently listening on i.e. 4040.
		If the theater is not actively listening, -1 is returned.
	*/
	public int getPort();
}
