/*
	Code generation for the following grammar segment:
	
	BehaviorBody :=
	"{" 
		(
			  Initializer 
			| NestedBehaviorDeclaration 
			| StateVariableDeclaration
			| ConstructorDeclaration
			| MethodDeclaration
		)* 
	"}"
*/

package salsac.definitions;

import salsac.SalsaCompiler;
import salsac.SalsaParser;
import salsac.SimpleNode;


public class BehaviorBody extends SimpleNode {
	
	public BehaviorBody(int id) 			{ super(id); }
	public BehaviorBody(SalsaParser p, int id)	{ super(p, id); }

	public String getPreCode()		{
		SalsaCompiler.indent++;
		String actorName = parent.getToken(1).image;

		String code = "{\n";
		code += SalsaCompiler.getIndent() + "public " + actorName + " selfReference;\n\n";

		/**
			Generate code for the constructors which get actor by name and get actor by reference.
		*/
		code += SalsaCompiler.getIndent() + "public " + actorName + "(UAN uan) { super(uan); }\n";
		code += SalsaCompiler.getIndent() + "public " + actorName + "(UAL ual) { super(ual); }\n\n";

		/**
			Create and send a message which will invoke act on the actor.
		*/
		code += SalsaCompiler.getIndent() + "public static void main(String args[]) {\n";
		code += SalsaCompiler.getIndent() + "\t" + actorName + " instance = new " + actorName + "();\n";
		code += SalsaCompiler.getIndent() + "\tMessage message = new Message( null, instance, \"act\", null );\n";
		code += SalsaCompiler.getIndent() + "\tObject[] _arguments = { args };\n";
		code += SalsaCompiler.getIndent() + "\tmessage.setArguments( _arguments );\n";
		code += SalsaCompiler.getIndent() + "\tinstance.send( message );\n";
		code += SalsaCompiler.getIndent() + "}\n\n";

		/**
			Generate code for the constructors.
		*/
		boolean hasBasic = false;

		if (children != null) {
			for (int i = 0; i < children.length; i++) {
				if (getChild(i) instanceof ConstructorDeclaration) {
					if ( getChild(i).getChild(1).children == null ) hasBasic = true;

					code += SalsaCompiler.getIndent() + "public " + actorName + getChild(i).getChild(1).getJavaCode() + " {\n";
					code += SalsaCompiler.getIndent() + "\tselfReference = this;\n";
					code += SalsaCompiler.getIndent() + "\tState state = new State(" + ((FormalParameters)getChild(i).getChild(1)).getParameterNames() + ");\n";
					code += SalsaCompiler.getIndent() + "\tServiceFactory.getTheater().setEntry(ual, state);\n";
					code += SalsaCompiler.getIndent() + "}\n";
				}
			}
		}
		if (!hasBasic) {
			code += SalsaCompiler.getIndent() + "public " + actorName + "() {\n";
			code += SalsaCompiler.getIndent() + "\tselfReference = this;\n";
			code += SalsaCompiler.getIndent() + "\tState state = new State();\n";
			code += SalsaCompiler.getIndent() + "\tServiceFactory.getTheater().setEntry(ual, state);\n";
			code += SalsaCompiler.getIndent() + "}\n";
		}


		/**
			Actually create the state.
			If this actor extends another actor, have it extend that actors state
		*/

		if ( ((BehaviorDeclaration)parent).extendsName != null) {
			String extendsName = ((BehaviorDeclaration)parent).extendsName;

			if (extendsName.equals("AutonomousActorReference")) {
				code += SalsaCompiler.getIndent() + "public class State extends AutonomousActor";
			} else {
				code += SalsaCompiler.getIndent() + "public class State extends " + ((BehaviorDeclaration)parent).extendsName + ".State";
			}
		} else {
			code += SalsaCompiler.getIndent() + "public class State extends Actor";
		}
		if ( ((BehaviorDeclaration)parent).implementNames.size() > 0) {
			code += " implements ";
			for (int i = 0; i < ((BehaviorDeclaration)parent).implementNames.size(); i++) {
				code += ((BehaviorDeclaration)parent).implementNames.get(i) + ".State";
				if (i+1 != ((BehaviorDeclaration)parent).implementNames.size()) code += ", ";
			}
		}
		code += " {\n";

		SalsaCompiler.indent++;
		code += SalsaCompiler.getIndent() + "protected " + actorName + " self = selfReference;\n";

		if (!hasBasic) {
			code += SalsaCompiler.getIndent() + "public State() {\n";
			code += SalsaCompiler.getIndent() + "\tself.ual = __ual;\n";
			code += SalsaCompiler.getIndent() + "}\n";
		}

		code += SalsaCompiler.getIndent() + "public void process(Message message) {\n";
		code += SalsaCompiler.getIndent() + "\tMethod match = getInvocationTarget(message);\n";
		code += SalsaCompiler.getIndent() + "\tObject token = null;\n";
		code += SalsaCompiler.getIndent() + "\tif (match != null) {\n";
		code += SalsaCompiler.getIndent() + "\t\ttry {\n";
		code += SalsaCompiler.getIndent() + "\t\t\ttoken = match.invoke(this, message.getArguments());\n";
		code += SalsaCompiler.getIndent() + "\t\t} catch (InvocationTargetException e) {\n";
		code += SalsaCompiler.getIndent() + "\t\t\tThrowable wrapped = e.getTargetException();\n";
		code += SalsaCompiler.getIndent() + "\t\t\tif ( !(wrapped instanceof CurrentContinuationException) ) {\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\tSystem.err.println(\"Message processing exception:\");\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\tif (message.getSender() != null) {\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\t\tSystem.err.println(\"\tSent by: \" + message.getSender().toString());\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\t} else System.err.println(\"\tSent by: unknown\");\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\tSystem.err.println(\"\tBy actor: \" + toString());\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\tSystem.err.println(\"\tInvocation of \" + message.getMethodName() + \" generated a \" + wrapped.getClass().getName() + \" exception: \" + wrapped.getMessage() );\n";
		code += SalsaCompiler.getIndent() + "\t\t\t}\n";
		code += SalsaCompiler.getIndent() + "\t\t} catch (IllegalAccessException e) {\n";
		code += SalsaCompiler.getIndent() + "\t\t\tSystem.err.println(\"Message processing exception:\");\n";
		code += SalsaCompiler.getIndent() + "\t\t\tif (message.getSender() != null) {\n";
		code += SalsaCompiler.getIndent() + "\t\t\t\tSystem.err.println(\"\tSent by: \" + message.getSender().toString());\n";
		code += SalsaCompiler.getIndent() + "\t\t\t} else System.err.println(\"\tSent by: unknown\");\n";
		code += SalsaCompiler.getIndent() + "\t\t\tSystem.err.println(\"\tBy actor: \" + toString());\n";
		code += SalsaCompiler.getIndent() + "\t\t\tSystem.err.println(\"\tIllegal access exception generated by the invocation of \" + message.getMethodName() + \": \" + e.getMessage() );\n";
		code += SalsaCompiler.getIndent() + "\t\t}\n";
		code += SalsaCompiler.getIndent() + "\t}\n";
		code += SalsaCompiler.getIndent() + "\ttransportService.send( messages );\n";
		code += SalsaCompiler.getIndent() + "\tmessages = new Vector();\n";
		code += SalsaCompiler.getIndent() + "\tif (currentContinuation != null) currentContinuation.resolve(message.getTarget(), token);\n";
		code += SalsaCompiler.getIndent() + "}\n";

		code += SalsaCompiler.getIndent() + "private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException {\n";
		code += SalsaCompiler.getIndent() + "\tself.uan = this.__uan;\n";
		code += SalsaCompiler.getIndent() + "\tself.ual = this.__ual;\n";
		code += SalsaCompiler.getIndent() + "\tout.defaultWriteObject();\n";
		code += SalsaCompiler.getIndent() + "\tout.flush();\n";
		code += SalsaCompiler.getIndent() + "}\n";

		code += SalsaCompiler.getIndent() + "private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, ClassNotFoundException {\n";
		code += SalsaCompiler.getIndent() + "\tin.defaultReadObject();\n";
//		code += SalsaCompiler.getIndent() + "\tstart();\n";
		code += SalsaCompiler.getIndent() + "}\n\n";

		return code;
	}

	public String getPostCode()	{
		SalsaCompiler.indent--;

		String code = SalsaCompiler.getIndent() + "}\n";
		code += "}";
	
		return code;
	}
	
	public String getChildCode() {
		String code = "";

		if (children != null) {		
			for (int i = 0; i < children.length; i++) {
				code += SalsaCompiler.getIndent() + getChild(i).getJavaCode();
			}
		}
		
		return code;
	}
}
