/*
	Code generation for the following grammar segment:
	
	MessageSend :=
		[PrimaryExpression "<-"] <IDENTIFIER> MessageArguments [":" MessageProperty]
*/

package salsac.definitions;

import salsac.*;

public class MessageSend extends SimpleNode {

	public MessageSend(int id) 			{ super(id); }
	public MessageSend(SalsaParser p, int id)	{ super(p, id); }

	public int getRequiredTokens() {
		int requiredTokens = 0;

		MessageArguments args = null;
		if (children[0] instanceof MessageArguments)	args = (MessageArguments)getChild(0);
		else 						args = (MessageArguments)getChild(1);

		int length = 0;
		if (args.children != null) length = args.children.length;
		for (int i = 0; i < length; i++) {
			String currentArgument = args.getChild(i).getJavaCode();

			if ( currentArgument.equals("token") ) {
				requiredTokens++;
			} else {
				String symbolType = SalsaCompiler.symbolTable.getSymbolType( currentArgument );

				if (symbolType != null && (symbolType.equals("next") || symbolType.equals("token")) ) {
					requiredTokens++;
				}
			}
		}

		return requiredTokens;
	}

	public String getArgumentsCode() {
		String code = "";
		if (children[0] instanceof MessageArguments) {
			code = "Object _arguments[] = { " + ((MessageArguments)getChild(0)).getMessageArguments("self") + " };\n";
		} else {
			code = "Object _arguments[] = { " + ((MessageArguments)getChild(1)).getMessageArguments(getChild(0).getJavaCode()) + " };\n";
		}		
		return code;
	}

	public String getMessageSendCode() {
		//new Message(<sender>, <target>, <message name>, <arguments>, <input token>, <output token>);
		String code = "";
		
		if (children[0] instanceof Expression) {
			code = "Message message = new Message( self, " + getChild(0).getJavaCode() + ", \"" + getToken(1).image + "\", ";
		} else {
			code = "Message message = new Message( self, self, \"" + getToken(0).image + "\", ";
		}
		
		String output = SalsaCompiler.symbolTable.getContinuationOutput();
		String joinOutput = SalsaCompiler.symbolTable.getJoinBlockOutput();

		String outputType = SalsaCompiler.symbolTable.getSymbolType(output);

		if (!joinOutput.equals("null") && outputType != null && outputType.equals("token") ) {
			System.err.println("Currently, named tokens are not allowed inside join blocks.");
			System.err.println("\tline: " + getToken(0).beginLine);
			System.exit(0);
		}
		if (output.equals("null")) output = joinOutput;
		if (!output.equals("null") && !output.equals("currentContinuation")) output += ".createInput()";

		code += output + " );\n";
		return code;
	}

	public String getInputToken() {
		String input = SalsaCompiler.symbolTable.getContinuationInput();
		if (!input.equals("null")) {
			boolean containsInput = false;
			if (children[0] instanceof MessageArguments) 	containsInput = ((MessageArguments)getChild(0)).containsInput(input);
			else 						containsInput = ((MessageArguments)getChild(1)).containsInput(input);

			if (containsInput) return "null";

			if (children[0] instanceof Expression) input += ".createOutput(" + getChild(0).getJavaCode() + ", message, -1)";
			else input += ".createOutput(self, message, -1)";
		}

		return input;
	}

	public String getProperty() {
		if (children.length == 3)							return getChild(2).getJavaCode();
		else if (children.length == 2 && getChild(0) instanceof MessageArguments)	return getChild(1).getJavaCode();

		return "";
	}

	public String getJavaCode() {
		String code = "";

		code += SalsaCompiler.getIndent() + "{\n";
		SalsaCompiler.indent++;
		code += SalsaCompiler.getIndent() + getMessageSendCode();
		code += SalsaCompiler.getIndent() + getArgumentsCode();
		code += SalsaCompiler.getIndent() + "message.setArguments( _arguments );\n";
		code += SalsaCompiler.getIndent() + "message.setRequiredTokens( " + getRequiredTokens() + " );\n";
		code += SalsaCompiler.getIndent() + "message.setInput( " + getInputToken() + " );\n";
		code += getProperty();

		code += SalsaCompiler.getIndent() + "messages.add( message );\n";

		SalsaCompiler.indent--;
		code += SalsaCompiler.getIndent() + "}\n";

		return code;
	}
}
