/*
 * Created on Sep 17, 2003
 *
 */

package wwc.messaging;

import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Vector;
import java.util.Hashtable;

import salsa.language.*;
import salsa.language.exceptions.*;
import salsa.naming.*;
import salsa.messaging.*;


/**
 * This class is the reference implementation of the Persistent MessagingService. The
 * service is implemented as a RMSP client.
 * 
 * This implementation uses the NamingService interface and the salsa.language.
 * ServiceFactory for UAN resolution.
 * 
 * @authors Travis Desell, Kaoutar El Maghraoui
 */
 
public class PMessenger implements TransportService {
	/**
	 * Constructor for Messenger.
	 */
	
	private static Hashtable inputStreams 	= new Hashtable();
	private static Hashtable outputStreams 	= new Hashtable();
	private static Hashtable sockets	= new Hashtable();

	
	
	public PMessenger() {
		super();
	}
	
	
	private ObjectInputStream  getInputStream(UAL target) throws UnknownHostException, IOException{
		
		ObjectInputStream  in = null;
		Socket socket = null;
		
		String hKey = target.getHost()+target.getPort();
		
		socket = (Socket)PMessenger.sockets.get(hKey);
		in = (ObjectInputStream) PMessenger.inputStreams.get(hKey);		
		
		if(socket == null || socket.isClosed() || !socket.isConnected())
		{
			
			//Create a new socket, either an exisiting socket has been disconnected or closed
			//or a new socket needs to be created since this is the first time we are contacting
	    		//this location

			try {
				socket = new Socket(target.getHost(), target.getPort());
				socket.setKeepAlive(true);
				socket.setTcpNoDelay(true);
			} catch (UnknownHostException e) {
				System.err.println("Unknown host: " + target.toString());
				System.err.println("\terror: " + e);
			} catch (IOException e) {
				System.err.println("Error in creating a socket for an output stream: " + e);
				System.err.println("\twith target: " + target.toString());
			
			} catch (Exception e) {
				System.err.println("Exception: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			
			PMessenger.sockets.put(hKey, socket);
			
			try {
				in = new ObjectInputStream(socket.getInputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			PMessenger.inputStreams.put(hKey, in);
		}		
		else if(in == null){
			
			try {
				
				in = new ObjectInputStream(socket.getInputStream());
				
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			
			PMessenger.inputStreams.put(hKey, in);

			
			
		}
						
		return in;
	}
	
	private ObjectOutputStream getOutputStream(UAL target) throws UnknownHostException, IOException{
		
		ObjectOutputStream  out = null;
		Socket socket = null;
		
		String hKey = target.getHost()+target.getPort();
		
		out = (ObjectOutputStream) PMessenger.outputStreams.get(hKey);
		socket = (Socket)PMessenger.sockets.get(hKey);
		
		if(socket == null || socket.isClosed() || !socket.isConnected())
		{
			//Create a new socket, either an exisiting socket has been disconnected or closed
			//or a new socket needs to be created since this is the first time we are contacting
			//this location
			 
			try {
				socket = new Socket(target.getHost(), target.getPort());
				socket.setKeepAlive(true);
				socket.setTcpNoDelay(true);
			} catch (UnknownHostException e) {
				System.err.println("Unknown host: " + target.toString());
				System.err.println("\terror: " + e);
			} catch (IOException e) {
				System.err.println("Error in creating a socket for an output stream: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			
			PMessenger.sockets.put(hKey, socket);
			
			try {
				out = new ObjectOutputStream(socket.getOutputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			PMessenger.outputStreams.put(hKey, out);
		}
		else if(out == null){
			try {
					out = new ObjectOutputStream(socket.getOutputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			PMessenger.outputStreams.put(hKey, out);		
		
		}
					
		return out;
	}

	public void migrate(Actor state, UAL target){
		boolean acked = false;
		
		//--Debug:System.out.println("\nMigrating  to ..."+target.getLocation()+"\n\n");
						
		try {
			ObjectOutputStream outStream = (ObjectOutputStream)getOutputStream(target);
	
			synchronized(outStream){
				outStream.writeObject( state );
				outStream.flush();
			}
		} catch (IOException e) {
			throw new SalsaWrapperException("Error migrating actor: ", e);
		}

		
		// Wait for the Acknowledgement that the actor is situated on the remote theater.
		try {
			ObjectInputStream inStream = (ObjectInputStream)getInputStream(target);
			
			while(!acked){
					
				Object received = inStream.readObject();
				if (received instanceof RMSPAcknowledgement){
					//--Debug:System.out.println("Received migration ack");
					acked = true;
				}
				else if (received instanceof Actor) {
					//Actor incomingActor = (Actor) received;
					//processActor(incomingActor, incoming);
					System.err.println("PMessenger: Received actor while expecting an ack");
				} else if (received instanceof Message) {  
					//Message message = (Message) received;								
					//processMessage(message);
					System.err.println("PMessenger: Received message while expecting an ack");			
				}
			}	
			//RMSPAcknowledgement ack = (RMSPAcknowledgement) inStream.readObject();
			

		} catch (Exception e) {
			throw new SalsaWrapperException( "Did not recieve RmspAcknowledgement. ", e);
		}


	}

	public void send(Message message, UAL target) {
		// Determine the destination of this message.

		if (target == null) {
			UAN uan = message.getTarget().getUAN();

			if (uan != null) {
				Object entry = ServiceFactory.getTheater().getEntry(uan);
				if ((entry instanceof UAL) || (entry == null)) {
					// Resolve the location.
					target = ServiceFactory.getNaming().resolve( uan );
					ServiceFactory.getTheater().setEntry(uan, target);
				} else {
					((Actor)entry).send(message);
					return;
				}
			} else {
				System.err.println("UAN and UAL of message are null, cannot send message");
			}
		}

		try {
			ObjectOutputStream outStream = getOutputStream(target);
			synchronized(outStream){
				outStream.writeObject( message );
				outStream.flush();
			}
		} catch (IOException e) {
			throw new SalsaWrapperException("Error sending message: ", e);
		}
	}

        public void send(Vector messages) {
                //used by the compiler to send a vector of messages generated by a message

                for (int i = 0; i < messages.size(); i++) {
                        final Message current = (Message)messages.get(i);

                        final ActorReference target = current.getTarget();
                        try {
                                if (current.getProperty() != null && current.getProperty().equals("delay")) {
                                        //This supports sending messages with the delay property
                                        Object[] parameters = current.getPropertyParameters();
                                        final long sleepTime = ((Integer)parameters[0]).intValue();

                                        Runnable delayThread = new Runnable() {
                                                public void run() {
                                                        try {
                                                                Thread.sleep(sleepTime);
                                                        } catch (Exception e) {
                                                                System.err.println("Error in delay thread for message: " + current.getName());
                                                        }
                                                        target.send(current);
                                                }
                                        };
                                        new Thread(delayThread, "Delay Thread").start();
                                } else {
                                        target.send( current );
                                }
                        } catch (Exception e) {
                                System.err.println("Transport service error: ");
                                System.err.println("\tSending message: " + current.getDescriptor());
                                if (current.getSender() != null) {
                                                System.err.println("\tBy actor: " + current.getSender().toString());
                                } else System.err.println("\tBy actor: unknown");
                                System.err.println("\tTo target: " + current.toString());
                                System.err.println("\tException: " + e);
                        }
                }
        }
}



